/****************************************************************
**
** Implementation of CurveParameters class, Qt tutorial 8
**
****************************************************************/

#include "testseries.h"
#include "CurveParameters.h"

#include <iostream>
#include <fstream>
#include <qpushbutton.h>
#include <qlcdnumber.h>
#include <qbuttongroup.h>
#include <qradiobutton.h>
#include <qlayout.h>
#include <qfiledialog.h>
#include <qapplication.h>
#include <qcombobox.h>
#include <qlabel.h>
#include <qgrid.h>

#ifndef PI
#define PI 3.1415926535897932384626433832795
#endif

CurveParameters::CurveParameters( QWidget *parent, const char *name )
        : QWidget( parent, name )
{
    QGridLayout *grid = new QGridLayout( this, 14, 1, 10, 10 );
    QButtonGroup* qbgrp = new QButtonGroup(2,Horizontal,"Rigid Motion",this,"move");
    btn1 = new QPushButton("<",qbgrp,"rotate_left");
    btn1->setAutoRepeat(TRUE);
    btn2 = new QPushButton(">",qbgrp,"rotate_right");
    btn2->setAutoRepeat(TRUE);
    QPushButton* btn_up = new QPushButton("Up",qbgrp,"up");
    btn_up->setAutoRepeat(TRUE);
    QPushButton* btn_right = new QPushButton("Right",qbgrp,"right");
    btn_right->setAutoRepeat(TRUE);
    QPushButton* btn_down = new QPushButton("Down",qbgrp,"down");
    btn_down->setAutoRepeat(TRUE);
    QPushButton* btn_left = new QPushButton("Left",qbgrp,"left");
    btn_left->setAutoRepeat(TRUE);

    connect(qbgrp,SIGNAL(pressed(int)),SLOT(move(int)));

    eps = 10.0;
    eps_prime = 20.0;

    EpsControl* qeps = new EpsControl("eps = ",10.0,0.0,100.0,0.5,this,"eps");
    connect(qeps,SIGNAL(valueChanged(double)),SLOT(setEps(double)));
    EpsControl* qeps_prime = new EpsControl("delta = ",20.0,0.0,100.0,0.5,this,"eps_prime");
    connect(qeps_prime,SIGNAL(valueChanged(double)),SLOT(setEps_prime(double)));

    curve_select = new QButtonGroup(1,Horizontal,"Curve",this,"curve_select");

    button_red = new QPushButton("",curve_select,"curve_select_red");
    button_red->setPalette( QPalette( QColor(255, 0, 0) ) );
    button_red->setToggleButton(TRUE);
    button_red->setDisabled(TRUE);
    connect(button_red,SIGNAL(pressed()),SLOT(select_red()));

    button_blue = new QPushButton("",curve_select,"curve_select_blue");
    button_blue->setPalette( QPalette( QColor(0, 0, 255) ) );
    button_blue->setToggleButton(TRUE);
    button_blue->toggle();
    button_blue->setDisabled(FALSE);
    connect(button_blue,SIGNAL(pressed()),SLOT(select_blue()));

    QLabel *method_label = new QLabel("Method:",this);
    QLabel *leash_label = new QLabel("Dog Leash:",this);

    QComboBox *select_method  = new QComboBox( this, "select_method" );
    select_method->insertItem("fast sampling");
    select_method->insertItem("standard sampling");
    select_method->insertItem("standard sampling discrete");
    select_method->insertItem("partial match");
    //select_method->insertItem("common subcurve match");
    select_method->insertItem("directed hausdorff match");
	//select_method->insertItem("exp-time sampling (be warned!)");

    connect(select_method,SIGNAL(activated(const QString&)),SLOT(method_changed(const QString&)));

    QHBox *reparam_box = new QHBox(this);

    QPushButton *prev_pair = new QPushButton( "<", reparam_box, "prev_pair" );
    prev_pair->setAutoRepeat(TRUE);
    QPushButton *next_pair = new QPushButton( ">", reparam_box, "next_pair" );
    next_pair->setAutoRepeat(TRUE);
    next_pair->setFont( QFont( "Arial", 8, QFont::Normal ) );
    prev_pair->setFont( QFont( "Arial", 8, QFont::Normal ) );

    connect( next_pair, SIGNAL(clicked()), this, SLOT(next_pair()) );
    connect( prev_pair, SIGNAL(clicked()), this, SLOT(prev_pair()) );

    QPushButton *match = new QPushButton( "Match!", this, "match" );
    match->setFont( QFont( "Arial", 12, QFont::Bold ) );
    connect(match,SIGNAL(released()),SLOT(match()));

    QPushButton *kill = new QPushButton( "Del Point", this, "kill" );
    kill->setFont( QFont( "Arial", 12, QFont::Bold ) );
    connect(kill,SIGNAL(pressed()),SLOT(del()));

    QPushButton *kill_all = new QPushButton( "Clear", this, "kill_all" );
    kill_all->setFont( QFont( "Arial", 12, QFont::Bold ) );
    connect(kill_all,SIGNAL(pressed()),SLOT(clear_curve()));

    qsave = new QPushButton( "Save", this, "save" );
    qsave->setFont( QFont( "Arial", 12, QFont::Bold ) );
    qsave->setToggleButton(FALSE);
    connect(qsave,SIGNAL(released()),SLOT(save()));

    qopen = new QPushButton( "Open", this, "open" );
    qopen->setFont( QFont( "Arial", 12, QFont::Bold ) );
    qopen->setToggleButton(FALSE);
    connect(qopen,SIGNAL(released()),SLOT(open()));

    QPushButton *testseries = new QPushButton( "Test Series", this, "testseries" );
    testseries->setFont( QFont( "Arial", 12, QFont::Bold ) );
    connect( testseries, SIGNAL(released()), SLOT(testseries()) );

    QPushButton *quit = new QPushButton( "Quit", this, "quit" );
    quit->setFont( QFont( "Arial", 12, QFont::Bold ) );
    connect( quit, SIGNAL(clicked()), qApp, SLOT(quit()) );

    grid->addWidget(curve_select,0,0);
    grid->addWidget(qbgrp,1,0);
    grid->addWidget(qeps,2,0);
    grid->addWidget(qeps_prime,3,0);
    grid->addWidget(method_label,4,0);
    grid->addWidget(select_method,5,0);
    grid->addWidget(leash_label,6,0);
    grid->addWidget(reparam_box,7,0);
    grid->addWidget(match,8,0);
    grid->addWidget(kill,9,0);
    grid->addWidget(kill_all,10,0);
    grid->addWidget(qopen,11,0);
    grid->addWidget(qsave,12,0);
    grid->addWidget(testseries,13,0);
    grid->addWidget(quit,14,0);

}

void CurveParameters::match()
{
    emit match_now(eps,eps_prime);
}

void CurveParameters::del()
{
    emit remove_last_point();
}

void CurveParameters::next_pair()
{
    emit show_next_pair();
}

void CurveParameters::prev_pair()
{
    emit show_prev_pair();
}

void CurveParameters::clear_curve()
{
    emit clear_active_curve();
}

void CurveParameters::setEps(double e)
{
    eps = e;
    emit epsChanged(e);
}

void CurveParameters::setEps_prime(double e)
{
    eps_prime = e;
    emit eps_primeChanged(e);
}

void CurveParameters::method_changed(const QString& mthd)
{
    emit change_method(mthd);
}

void CurveParameters::select_red()
{
    if (!button_red->isEnabled())
        return;
    button_blue->toggle();
    button_red->toggle();
    button_red->setDisabled(TRUE);
    button_blue->setDisabled(FALSE);
    emit colorChanged(0);
}

void CurveParameters::open()
{
    repaint();
    QString s( QFileDialog::getOpenFileName( QString::null, "Curves (*.rsc)", this ) );
    repaint();
    if ( s.isEmpty() )
    {
        repaint();
        return;
    }
    repaint();
    emit open_file(s);
}

void CurveParameters::testseries()
{
    repaint();
    QString s( QFileDialog::getOpenFileName( QString::null, "Curves (*.ts)", this ) );
    repaint();
    if ( s.isEmpty() )
    {
        repaint();
        return;
    }
    repaint();

    my_testseries(s);
}

void CurveParameters::save()
{
    repaint();
    QString s( QFileDialog::getSaveFileName( QString::null, "Curves (*.rsc)", this ) );
    repaint();
    if ( s.isEmpty() )
    {
        repaint();
        return;
    }
    repaint();
    emit save_file(s);
}

void CurveParameters::select_blue()
{
    if (!button_blue->isEnabled())
        return;
    button_red->toggle();
    button_blue->toggle();
    button_blue->setDisabled(TRUE);
    button_red->setDisabled(FALSE);
    emit colorChanged(1);
}

void CurveParameters::move(int type)
{
    if (type==0)
    {
        emit rigid_motion(Motion(sin(PI/36.0),cos(PI/36.0),0,0));
    } else if (type==1) {
        emit rigid_motion(Motion(sin(-PI/36.0),cos(-PI/36.0),0,0));
    } else if (type==2) {
        emit rigid_motion(Motion(0.0,1.0,0.0,-5.0));
    } else if (type==3) {
        emit rigid_motion(Motion(0.0,1.0,5.0,0.0));
    } else if (type==4) {
        emit rigid_motion(Motion(0.0,1.0,0.0,5.0));
    } else if (type==5) {
        emit rigid_motion(Motion(0.0,1.0,-5.0,0.0));
    }
}

void CurveParameters::rotate_left()
{
}

void CurveParameters::rotate_right()
{
}

