#ifndef RESCUE__LCSC_MATCH_GRAPH_H
#define RESCUE__LCSC_MATCH_GRAPH_H

#include <vector>
#include <list>
#include <algorithm>
#include "rescutil.h"
#include "compact_interval_border.h"


namespace rescue
{


/** The extended free space graph is an essential part of the improved matching-algorithm used by the class
\c fast_sampling_matcher. The class \c lcsc_match_graph supplies all neccessary operations for
such a graph, in particular finding certain maximal paths and dynamically extending the active
configuration of edges.
*/
// ==========================================================================================
// ==========================================================================================
template<class transporter_edge>
class lcsc_match_graph 
{

public:

    typedef transporter_edge transporter_edge_class;
    typedef transporter_edge::index_t index_t;
    typedef transporter_edge::arithmetic_t arithmetic;
    typedef transporter_edge::interval_border interval_border;
    typedef transporter_edge::border_iterator edge_border_iterator;
    typedef transporter_edge::point_t point;
    typedef transporter_edge::motion motion;

    typedef arithmetic arithmetic_t;

	typedef std::pair<point,point> point_pair;
    
    /** Type by which the interval borders are described.
    */
    typedef transporter_edge::border_value border_value;

    /** The matching task only yields discretized curves through the approximate free
    space. Thus, the points in free space can be described by integers (represented by
    the type \c index_t).
    */
    typedef std::pair<index_t,index_t> discrete_free_space_point;

    /** A \c discrete_reparametrization_pair is a sequence of (discrete) points contained
    in the (approxiamte) free space.
    */
    typedef std::list<discrete_free_space_point> discrete_reparametrization_pair;

    
    typedef std::vector<interval_border> border_array;
    typedef std::list<interval_border> border_list;
    typedef border_array::iterator border_iterator;

    /** Default constructor. Initializes an empty graph; for initialization, use
    \c build() .
    */
    lcsc_match_graph();

	/** Default destructor; calls \c clear() in order to deallocate all reserved memory.
    */
    ~lcsc_match_graph();

	/** Constructs the free space graph of the polygonal curves \c P and \c Q of lengths
    \c m and \c n , respectively. \c eps specifies the fault tolerance and \c delta 
    specifies the sampling density and thus the quality of approximation.
    */
    void build(point* P,point* Q,index_t m_P,index_t n_Q, index_t k_P, index_t l_Q,
		const arithmetic& eps, const arithmetic& delta);

    /** Updates reachability etc. by changing all properties affected by
    (de-)activating the edge indicated by \c ib.
    */
    arithmetic update(const interval_border& update_borders);

    /** Determines whether the maximum i sucht that there is a path from (0,0) to some (i,j) 
    (i.e., a flow) along the currently active edges through the graph. 
    Each instance of the flow problem takes O(mn) time (using dynamic programming).    
    */
	arithmetic find_path();

    /** Returns an element \c g of the transformation group with 
    the property that d_F(P,gQ)<=2(eps+delta), if such \c g exists.
    The output refers to the last call of \c match() .
    If no such \c g exists or if \c match() has not yet been called, the 
    output is not specified.
    */
    motion get_match();

    /** Returns a pair (\c kappa, \c lambda) containing the discretized reparametrizations
    that deschribe a monotonic curve in the 2(\c eps + \c delta)-free space.
    */
    bool get_reparametrizations(discrete_reparametrization_pair& kappa_lambda);

	/** Deallocates all memory reserved by the \c lcsc_match_graph object.
    */
    void clear();

private:

    /** Counts number of active edges in each column of the graph.
    This information is set and used for testing the suitability of the active edge
    configuration in \c update().
    */
    index_t* columns;

    /** Counts number of active edges in each row of the graph.
    This information is set and used for testing the suitability of the active edge
    configuration in \c update().
    */
    index_t* rows;
    
    /** Counts number columns having at least one active edge.
    This information is set and used for testing the suitability of the active edge
    configuration in \c update().
    */
    index_t active_columns;
    
    /** Counts number rows having at least one active edge.
    This information is set and used for testing the suitability of the active edge
    configuration in \c update().
    */
    index_t active_rows;

    /** Denotes the type of interval border encountered during the last call of
    \c update(); used for testing suitability of the active edge configuration.
    */
    border_type prev_border_type;

    /** The width of the graph, i.e. the length of the polygonal curve \c P
    specified in \c build() . Needs to be public for internal reasons; other
    classes should not access \c m anyway. (I know this needs to be fixed...)
    */
    index_t m;
	
	/** The height of the graph, i.e. the length of the polygonal curve \c Q
    specified in \c build() . Needs to be public for internal reasons; other
    classes should not access \c n anyway. (I know this needs to be fixed...)
    */
    index_t n;

    /** For common subcurve matchings, q_l is identified with some p_k. The value \c k is what
    is stored in this variable. It is updated in the most recent call of the \c build() 
    method by the value \c k_P that is passed to \c build().
    */
    index_t k;
	
    /** For common subcurve matchings, q_l is identified with some p_k. The value \c l is what
    is stored in this variable. It is updated in the most recent call of the \c build() 
    method by the value \c l_Q that is passed to \c build().
    */
    index_t l;
	
    
	/** The set of all edges of the graph. Needs to be public for internal reasons; other
    classes should not access \c edges anyway. (I know this needs to be fixed...)
    */
    transporter_edge*** edges; 
    
	/** The set of all active edges of the graph. 
    */
    bool*** edge_activated;
    
	/** Keeps track of the \c update_step during which each edge has been activated.
    */
    index_t*** activation_step;

    /** Reachability matrix of the active configuration of edges.
    */
    bool** reachable;

    /** Used for storing the Euclidean length of the curve <p_0,...p_i> in \c length_P[i].

    */
    arithmetic* length_P;

    /** Counts how often \c update(interval_bound_list) has been called.
    */
    index_t update_step;

    /** The reconstruction of a match of a match from a successful path through the graph
    requires a pair of reference points (usually p_0 and q_0 or points close to p_0 or 
    q_0). Thus, such a pair is stored together with the graph.
    */
    point_pair reference_points;

    /** Contains an interval border that represents a valid match, i.e. a maximal path
    through the graph with a non-empty intersection of the edges along the path.

    In combination with the \c reference_points, the \c match_pos allows a transformation
    to be computed that indicates a match between the curves \c P and \c Q.
    */
    border_value match_pos;

    /** Indicates whether the last instance of find_path did find maximal path through 
    the graph.
    */
    bool match_found;

    /** Indicates in which column the longest path determined by \c find_path() is terminated.
    */
    index_t match_column;

    /** Indicates in which row the longest path determined by \c find_path() is terminated.
    */
    index_t match_row;

    /** Indicates the length (in terms of the corresponding subcurve of P) of the longest path 
    determined by \c find_path().
    */
    arithmetic match_length;

    /** List of all (left and right) interval borders of the intervals assigned to the
    edges of the graph. This list is maintained temporarily during \c build() and
    is then maintained in the (sorted) vector \c interval_borders.
    */
    border_list interval_borders_list;

    /** Sorted vector of all (left and right) interval borders of the intervals assigned to the
    edges of the graph. This vector is available after \c build() has finished; the sorted
    vector is built from the temporary list \c interval_borders_list.
    */
    border_array interval_borders;

    /** Extracts all interval borders belonging to the interval \c E and inserts these
    interval borders in the \c interval_borders list.
    */
    void extract_borders(transporter_edge& E, const index_t& i, const index_t& j,
        graph_direction dir);

    /** Sorts all interval borders extracted by \c extract_borders()
    */
    void sort_borders();

public:

    /** This iterator allows to iterate over all interval borders of the edges assigned to
    the graph. The iteration is performed in ascending order.
    */
    border_iterator border_begin();

    /** See \c border_begin().
    */
	border_iterator border_end();

};

// ==========================================================================================
// ==========================================================================================




// ==========================================================================================
template<class transporter_edge>
lcsc_match_graph<transporter_edge>::
	lcsc_match_graph()
// ==========================================================================================
{
	m=0;
	n=0;
	edges = NULL;
    rows = NULL;
    columns = NULL;
    length_P = NULL;
    match_found = false;
}

// ==========================================================================================
template<class transporter_edge>
lcsc_match_graph<transporter_edge>::
	~lcsc_match_graph()
// ==========================================================================================
{
	clear();
}

// ==========================================================================================
template<class transporter_edge>
lcsc_match_graph<transporter_edge>::border_iterator
lcsc_match_graph<transporter_edge>::border_begin()
// ==========================================================================================
{
    return interval_borders.begin();
}

// ==========================================================================================
template<class transporter_edge>
lcsc_match_graph<transporter_edge>::border_iterator
lcsc_match_graph<transporter_edge>::border_end()
// ==========================================================================================
{
    return interval_borders.end();
}

// ==========================================================================================
template<class transporter_edge>
void lcsc_match_graph<transporter_edge>::
	clear()
// ==========================================================================================
{

    if (rows!=NULL)
        delete[] rows;
    if (columns!=NULL)
        delete[] columns;
    if (length_P!=NULL)
        delete[] length_P;
    rows = NULL;
    columns = NULL;
    length_P = NULL;

    if (edges!=NULL)
    {
	    for (index_t i=0; i<=m; i++)
	    {
		    for (index_t j=0; j<=n; j++)
		    {
			    delete[] edges[i][j];
			    delete[] edge_activated[i][j];
			    delete[] activation_step[i][j];
		    }
		    delete[] edges[i];
			delete[] edge_activated[i];
			delete[] activation_step[i];
			delete[] reachable[i];
	    }
	    delete[] edges;
		delete[] edge_activated;
		delete[] activation_step;
		delete[] reachable;

	    edges = NULL;
		edge_activated = NULL;
		activation_step = NULL;
		reachable = NULL;
    }

    m = 0;
    n = 0;
    match_found = false;

}

// ==========================================================================================
template<class transporter_edge>
void lcsc_match_graph<transporter_edge>::
	extract_borders(transporter_edge& E,
    const index_t& i, const index_t& j, graph_direction dir)
// ==========================================================================================
{

    edge_border_iterator r_it, r_end, l_it, l_end;

    r_it = E.right_borders_begin();
    r_end = E.right_borders_end();
    l_it = E.left_borders_begin();
    l_end = E.left_borders_end();

    for (; l_it!=l_end; l_it++)
    {
        interval_borders_list.push_back(interval_border(*l_it,LEFT_BORDER,i,j,dir));
    }

    for (; r_it!=r_end; r_it++)
    {
        interval_borders_list.push_back(interval_border(*r_it,RIGHT_BORDER,i,j,dir));
    }

}

// ==========================================================================================
template<class transporter_edge>
void lcsc_match_graph<transporter_edge>::
	sort_borders()
// ==========================================================================================
{
    
    // assign sufficient space to border_intervals
    index_t ii = interval_borders_list.size();
    interval_borders = border_array(interval_borders_list.size());

    // copy from border_interval_list to border_intervals
    std::copy(
        interval_borders_list.begin(),interval_borders_list.end(),
        interval_borders.begin());

    // sort border_intervals
    std::sort(interval_borders.begin(),interval_borders.end());

    // the unsorted list is needed no mo'.
    interval_borders_list.clear();

}

// ==========================================================================================
template<class transporter_edge>
void lcsc_match_graph<transporter_edge>::
	build(point* P,point* Q,index_t m_P,index_t n_Q, index_t k_P, index_t l_Q,
	const arithmetic& eps, const arithmetic& delta)
// ==========================================================================================
{

	clear();

	m = m_P;
	n = n_Q;
    k = k_P;
    l = l_Q;

    update_step = 0;

    // initialize the counters that help to determine the suitability
    // of the currently active intervals.
    prev_border_type = LEFT_BORDER;
    index_t i;
    columns = new index_t[m+1];
    rows = new index_t[n+1];
    for (i=0; i<=m; i++)
        columns[i] = 0;
    for (i=0; i<=n; i++)
        rows[i] = 0;
    active_columns = 0;
    active_rows = 0;

    length_P = new arithmetic[m+1];
	edges = new transporter_edge**[m+1];
	edge_activated = new bool**[m+1];
	activation_step = new index_t**[m+1];
    reachable = new bool*[m+1];

	reference_points = point_pair(P[k],Q[l]);
	
    length_P[0] = 0.0;
	for (i=0; i<=m; i++)
	{
		edges[i] = new transporter_edge*[n+1];
		edge_activated[i] = new bool*[n+1];
		activation_step[i] = new index_t*[n+1];
        reachable[i] = new bool[n+1];

        index_t i_1 = (i+k)%(m+1);
        index_t i_2 = (i+k+1)%(m+1);

        if (i<m)
            length_P[i+1] = length_P[i] + P[i].distance(P[i_2]);

        for (index_t j=0; j<n; j++)
		{
		
            index_t j_1 = (j+l)%(n+1);
            index_t j_2 = (j+l+1)%(n+1);

            edges[i][j] = new transporter_edge[3];
			edge_activated[i][j] = new bool[3];
			activation_step[i][j] = new index_t[3];

			edges[i][j][horizontal].assign_transporter
				(point_pair(P[i_1],P[i_2]),point_pair(Q[j_1],Q[j_1]),reference_points,eps,delta);
			edges[i][j][diagonal].assign_transporter
				(point_pair(P[i_1],P[i_2]),point_pair(Q[j_1],Q[j_2]),reference_points,eps,delta);
			edges[i][j][vertical].assign_transporter
				(point_pair(P[i_1],P[i_1]),point_pair(Q[j_1],Q[j_2]),reference_points,eps,delta);
            
            // extract interval borders from the three edges just created            
            extract_borders(edges[i][j][horizontal],i,j,horizontal);
            extract_borders(edges[i][j][diagonal],i,j,diagonal);
            extract_borders(edges[i][j][vertical],i,j,vertical);

            // initially, each edge is inactive (until it is activated in an update()-step)
            edge_activated[i][j][horizontal] = false;
            edge_activated[i][j][diagonal] = false;
            edge_activated[i][j][vertical] = false;

            // activation_step 0 means that the edge has never been activated.
            activation_step[i][j][horizontal] = 0;
            activation_step[i][j][diagonal] = 0;
            activation_step[i][j][vertical] = 0;

            // since all edges are inactive, no vertex is reachable
            // (except for the source (0,0), but we'll take care of that later...).
            reachable[i][j] = false;

		}

        reachable[i][n] = false;
		edges[i][n] = new transporter_edge[3];
		edge_activated[i][n] = new bool[3];
		activation_step[i][n] = new index_t[3];

		edges[i][n][horizontal].assign_transporter
				(point_pair(P[i_1],P[i_2]),point_pair(Q[n],Q[n]),reference_points,eps,delta);
        extract_borders(edges[i][n][horizontal],i,n,horizontal);
        edge_activated[i][n][horizontal] = false;
        activation_step[i][n][horizontal] = 0;

	}

    // Before we forget: the graph's source is reachable
    reachable[0][0] = true;

    // sort the extracted interval borders
    sort_borders();
		
}

// ==========================================================================================
template<class transporter_edge>
lcsc_match_graph<transporter_edge>::arithmetic lcsc_match_graph<transporter_edge>::
    update(const interval_border& ib)
// ==========================================================================================
{

	index_t i = ib.col();
    index_t j = ib.row();
    match_pos = ib.val();

    match_length = 0.0;

    // update suitability-data of currently active intervals
    if (ib.type()==LEFT_BORDER)
	    // toggle edge activation
	    edge_activated[i][j][ib.dir()] = true;

    if (ib.type()==RIGHT_BORDER && prev_border_type==LEFT_BORDER)
    {
        // the maximum match_length is updated in find_path; the overall
        // maximum found is to be returned later.
        arithmetic sc_length = find_path();
        if (sc_length>match_length)
        {
            match_length  = sc_length;
            match_pos = ib.val();        
        }
    }

    prev_border_type = ib.type();

    // The suitability-information for right interval broders
    // is updated after the path-update, because the border itself is still contained
    // in the intersection.
    if (ib.type()==RIGHT_BORDER)
    	// toggle edge activation
		edge_activated[i][j][ib.dir()] = false;
	
	return match_length;

}


// ==========================================================================================
template<class transporter_edge>
lcsc_match_graph<transporter_edge>::arithmetic lcsc_match_graph<transporter_edge>::
	find_path()
// ==========================================================================================
{
	bool* last_column;
	bool* curr_column;

    // use dynamic programming for updating the graph.	
	bool path_found;
    match_found = false;

    bool b_v, b_h, b_d;

	// initialize last_row
	last_column = reachable[0];

	last_column[0] = 0.0;
	for (index_t j=1; j<=n; j++)
		last_column[j] = 0.0;

	// in case m==0, curr_column[m][n] needs to specified...
	curr_column = last_column;
	index_t i=1;

	for (; i<=m-k; i++)
	{
		index_t prev_col = i-1;
		index_t col = i;
		last_column = reachable[prev_col];
		curr_column = reachable[col];

		path_found = false;
        curr_column[0] = true;

        b_h = edge_activated[prev_col][0][horizontal];
		if (last_column[0] && b_h)
		{
			curr_column[0] = true;
			path_found = true;
		}

		for (index_t j=1; j<=n-l; j++)
		{
			curr_column[j] = false;

			// if-clause slightly optimized "by hand"
			b_d = edge_activated[prev_col][j-1][diagonal];                    
			if (last_column[j-1] && b_d)
			{
                path_found = true;
				curr_column[j] = true;
			} else {
				b_v = edge_activated[col][j-1][vertical];
				if (curr_column[j-1] && b_v)
				{
                    path_found = true;
					curr_column[j] = true;
				} else {
					b_h = edge_activated[prev_col][j][horizontal];
    				if (last_column[j] && b_h)
                    {
                        path_found = true;
						curr_column[j] = true;
                    }
				}
			}
		}
		if (!path_found)
		{
            // The subcurves of P and Q could be matched up to the vertex p_{i-1}
            match_column = i;
            match_row = j;
            return length_P[i-1];
		}
	}

    // if we made it down here, the whole curve P has been matched!
    // Congrats!
    match_found = true;
    match_column = m;
    match_row = n;
	return length_P[m];

}


// ==========================================================================================
template<class transporter_edge>
lcsc_match_graph<transporter_edge>::motion 
    lcsc_match_graph<transporter_edge>::get_match()
// ==========================================================================================
{

    if (m==0 && n==0)
    {
        transporter_edge te;
        te.assign_transporter
            (point_pair(reference_points.first,reference_points.first),
             point_pair(reference_points.second,reference_points.second),
             reference_points,1.0,1.0);
        return transporter_edge::compute_match(*(te.left_borders_begin()),reference_points);

    }

    return transporter_edge::compute_match(match_pos,reference_points);

}

// ==========================================================================================
template<class transporter_edge>
bool lcsc_match_graph<transporter_edge>::get_reparametrizations
    (lcsc_match_graph<transporter_edge>::discrete_reparametrization_pair& kappa_lambda)
// ==========================================================================================
{

    kappa_lambda.clear();
    kappa_lambda.push_front(discrete_free_space_point(k,l));
    kappa_lambda.push_front(discrete_free_space_point
		((k+match_column)%(m+1),(l+match_row)%(m+1)));    
    return true;

}


// end of namespace rescue
}


#endif
