#ifndef RESCUE__LCSC_SAMPLING_MATCHER_H
#define RESCUE__LCSC_SAMPLING_MATCHER_H

#include "polygonal_curve.h"
#include "sampling_matcher.h"
#include "rescutil.h"

namespace rescue
{

/** An \c lcsc_sampling_matcher allows to approximately find longest common subcurves with 
respect to the Frechet distance by approximately solving the emptyness problem for the set 
of all Frechet matches.

The algorithm used is based on \c delta - samplings of polygonal curves.
*/

// ==========================================================================================
// ==========================================================================================
template<class fs_graph>
class lcsc_sampling_matcher : public sampling_matcher<fs_graph>
{

public:

    lcsc_sampling_matcher() 
        : sampling_matcher() {};

    lcsc_sampling_matcher(curve& P, curve& Q, const arithmetic& ee, const arithmetic& dd)
        : sampling_matcher<fs_graph>(P,Q,ee,dd) {};

    ~lcsc_sampling_matcher();

    /** Runs the matching algorithm, taking O(m^3n^2) time if both \c P and \c Q are
    \c delta - sampled.
    */
    bool match();

    /** Deallocates all memory used by the \c lcsc_sampling_matcher object.
    */
    void clear();

protected:

    arithmetic best_match;
};
// ==========================================================================================
// ==========================================================================================

// ==========================================================================================
template<class fs_graph>
bool lcsc_sampling_matcher<fs_graph>::match()
// ==========================================================================================
{

    if (m<0 || n<0)
        return false;

    Gamma.clear();

    best_match = 0.0;

    for (index_t k=0; k<=m; k++)
    {
        for (index_t l=0; l<=n; l++)
        {
        
	        Gamma.build(P,Q,m,n,k,l,eps,delta);

            // We treat the case that both P and Q are points as an exception
            if (m==0 && n==0)
            {
                match_motion = Gamma.get_match();
                return true;
            }

	        fs_graph::border_iterator b_it = Gamma.border_begin();
	        fs_graph::border_iterator b_end = Gamma.border_end();

	        for (; b_it!=b_end; b_it++)
	        {
                arithmetic current_match = Gamma.update(*b_it);
		        if (current_match>best_match)
                {
                    best_match = current_match;
                    match_motion = Gamma.get_match();
		            parametrizations_available 
						= Gamma.get_reparametrizations(match_kappa_lambda);
                }

	        }
        }
    }

    Gamma.clear();

    return true;

}


// ==========================================================================================
template<class fs_graph>
void lcsc_sampling_matcher<fs_graph>::clear()
// ==========================================================================================
{
    sampling_matcher<fs_graph>::clear();
}
    
// ==========================================================================================
template<class fs_graph>
lcsc_sampling_matcher<fs_graph>::~lcsc_sampling_matcher()
// ==========================================================================================
{
    clear();
}
    



// ==========================================================================================
// ==========================================================================================

// end of namespace rescue
}

#endif
