#ifndef RESCUE__PARTIAL_MATCH_GRAPH_H
#define RESCUE__PARTIAL_MATCH_GRAPH_H

#include <vector>
#include <list>
#include <algorithm>
#include "rescutil.h"
#include "compact_interval_border.h"


namespace rescue
{


/** The extended free space graph is an essential part of the improved matching-algorithm used by the class
\c fast_sampling_matcher. The class \c partial_match_graph supplies all neccessary operations for
such a graph, in particular finding certain maximal paths and dynamically extending the active
configuration of edges.
*/
// ==========================================================================================
// ==========================================================================================
template<class transporter_edge>
class partial_match_graph 
{

public:

    typedef transporter_edge transporter_edge_class;
    typedef transporter_edge::index_t index_t;
    typedef transporter_edge::arithmetic_t arithmetic;
    typedef transporter_edge::interval_border interval_border;
    typedef transporter_edge::border_iterator edge_border_iterator;
    typedef transporter_edge::point_t point;
    typedef transporter_edge::motion motion;

    typedef arithmetic arithmetic_t;

	typedef std::pair<point,point> point_pair;
    
    /** Type by which the interval borders are described.
    */
    typedef transporter_edge::border_value border_value;

    /** The matching task only yields discretized curves through the approximate free
    space. Thus, the points in free space can be described by integers (represented by
    the type \c index_t).
    */
    typedef std::pair<index_t,index_t> discrete_free_space_point;

    /** A \c discrete_reparametrization_pair is a sequence of (discrete) points contained
    in the (approxiamte) free space.
    */
    typedef std::list<discrete_free_space_point> discrete_reparametrization_pair;

    
    typedef std::vector<interval_border> border_array;
    typedef std::list<interval_border> border_list;
    typedef border_array::iterator border_iterator;

    /** Default constructor. Initializes an empty graph; for initialization, use
    \c build() .
    */
    partial_match_graph();

	/** Default destructor; calls \c clear() in order to deallocate all reserved memory.
    */
    ~partial_match_graph();

	/** Constructs the free space graph of the polygonal curves \c P and \c Q of lengths
    \c m and \c n , respectively. \c eps specifies the fault tolerance and \c delta 
    specifies the sampling density and thus the quality of approximation.
    */
    void build(point* P,point* Q,index_t m_P,index_t n_Q, index_t k_P,
		const arithmetic& eps, const arithmetic& delta);

    /** Updates reachability etc. by changing all properties affected by
    (de-)activating the edge indicated by \c ib.
    */
    bool update(const interval_border& update_borders);

    /** Determines whether there is a path from (0,0) to (m,n) (i.e., a flow) along the 
    currently active edges through the graph. Each instance of the flow problem takes 
    O(mn) time (using dynamic programming).    
    */
	bool find_path();

    /** Returns an element \c g of the transformation group with 
    the property that d_F(P,gQ)<=2(eps+delta), if such \c g exists.
    The output refers to the last call of \c match() .
    If no such \c g exists or if \c match() has not yet been called, the 
    output is not specified.
    */
    motion get_match();

    /** Returns a pair (\c kappa, \c lambda) containing the discretized reparametrizations
    that deschribe a monotonic curve in the 2(\c eps + \c delta)-free space.
    */
    bool get_reparametrizations(discrete_reparametrization_pair& kappa_lambda);

	/** Deallocates all memory reserved by the \c partial_match_graph object.
    */
    void clear();

private:

    /** Counts number of active edges in each column of the graph.
    This information is set and used for testing the suitability of the active edge
    configuration in \c update().
    */
    index_t* columns;

    /** Counts number of active edges in each row of the graph.
    This information is set and used for testing the suitability of the active edge
    configuration in \c update().
    */
    index_t* rows;
    
    /** Counts number columns having at least one active edge.
    This information is set and used for testing the suitability of the active edge
    configuration in \c update().
    */
    index_t active_columns;
    
    /** Counts number rows having at least one active edge.
    This information is set and used for testing the suitability of the active edge
    configuration in \c update().
    */
    index_t active_rows;

    /** Denotes the type of interval border encountered during the last call of
    \c update(); used for testing suitability of the active edge configuration.
    */
    border_type prev_border_type;

    /** The width of the graph, i.e. the length of the polygonal curve \c P
    specified in \c build() . Needs to be public for internal reasons; other
    classes should not access \c m anyway. (I know this needs to be fixed...)
    */
    index_t m;
	
	/** The height of the graph, i.e. the length of the polygonal curve \c Q
    specified in \c build() . Needs to be public for internal reasons; other
    classes should not access \c n anyway. (I know this needs to be fixed...)
    */
    index_t n;

    /** For partial matchings, q_0 is identified with some p_k. This value \c k is what
    is stored in this variable. It is updated in the most recent call of the \c build() 
    method by the value \c k_P that is passed to \c build().
    */
    index_t k;
	
    
	/** The set of all edges of the graph. Needs to be public for internal reasons; other
    classes should not access \c edges anyway. (I know this needs to be fixed...)
    */
    transporter_edge*** edges; 
    
	/** The set of all active edges of the graph. 
    */
    bool*** edge_activated;
    
	/** Keeps track of the \c update_step during which each edge has been activated.
    */
    index_t*** activation_step;

    /** Reachability matrix of the active configuration of edges.
    */
    bool** reachable;

    /** Counts how often \c update(interval_bound_list) has been called.
    */
    index_t update_step;

    /** The reconstruction of a match of a match from a successful path through the graph
    requires a pair of reference points (usually p_0 and q_0 or points close to p_0 or 
    q_0). Thus, such a pair is stored together with the graph.
    */
    point_pair reference_points;

    /** Contains an interval border that represents a valid match, i.e. a maximal path
    through the graph with a non-empty intersection of the edges along the path.

    In combination with the \c reference_points, the \c match_pos allows a transformation
    to be computed that indicates a match between the curves \c P and \c Q.
    */
    border_value match_pos;

    /** Indicates whether the last instance of find_path did find maximal path through 
    the graph.
    */
    bool match_found;

    /** Indicates in which column the \c find_path() procedure found a matching path. This
    column represents the end of the path through the free space that proves the existence
    of a match. The value \c match_column is updated whenever \c find_path() returns \c true.
    \c match_value is used for reconstructing reparametrizations.
    */
    index_t match_column;

    /** List of all (left and right) interval borders of the intervals assigned to the
    edges of the graph. This list is maintained temporarily during \c build() and
    is then maintained in the (sorted) vector \c interval_borders.
    */
    border_list interval_borders_list;

    /** Sorted vector of all (left and right) interval borders of the intervals assigned to the
    edges of the graph. This vector is available after \c build() has finished; the sorted
    vector is built from the temporary list \c interval_borders_list.
    */
    border_array interval_borders;

    /** Extracts all interval borders belonging to the interval \c E and inserts these
    interval borders in the \c interval_borders list.
    */
    void extract_borders(transporter_edge& E, const index_t& i, const index_t& j,
        graph_direction dir);

    /** Sorts all interval borders extracted by \c extract_borders()
    */
    void sort_borders();

public:

    /** This iterator allows to iterate over all interval borders of the edges assigned to
    the graph. The iteration is performed in ascending order.
    */
    border_iterator border_begin();

    /** See \c border_begin().
    */
	border_iterator border_end();

};

// ==========================================================================================
// ==========================================================================================




// ==========================================================================================
template<class transporter_edge>
partial_match_graph<transporter_edge>::
	partial_match_graph()
// ==========================================================================================
{
	m=0;
	n=0;
	edges = NULL;
    rows = NULL;
    columns = NULL;
    match_found = false;
}

// ==========================================================================================
template<class transporter_edge>
partial_match_graph<transporter_edge>::
	~partial_match_graph()
// ==========================================================================================
{
	clear();
}

// ==========================================================================================
template<class transporter_edge>
partial_match_graph<transporter_edge>::border_iterator
partial_match_graph<transporter_edge>::border_begin()
// ==========================================================================================
{
    return interval_borders.begin();
}

// ==========================================================================================
template<class transporter_edge>
partial_match_graph<transporter_edge>::border_iterator
partial_match_graph<transporter_edge>::border_end()
// ==========================================================================================
{
    return interval_borders.end();
}

// ==========================================================================================
template<class transporter_edge>
void partial_match_graph<transporter_edge>::
	clear()
// ==========================================================================================
{

    if (rows!=NULL)
        delete[] rows;
    if (columns!=NULL)
        delete[] columns;
    rows = NULL;
    columns = NULL;

    if (edges!=NULL)
    {
	    for (index_t i=0; i<=m; i++)
	    {
		    for (index_t j=0; j<=n; j++)
		    {
			    delete[] edges[i][j];
			    delete[] edge_activated[i][j];
			    delete[] activation_step[i][j];
		    }
		    delete[] edges[i];
			delete[] edge_activated[i];
			delete[] activation_step[i];
			delete[] reachable[i];
	    }
	    delete[] edges;
		delete[] edge_activated;
		delete[] activation_step;
		delete[] reachable;

	    edges = NULL;
		edge_activated = NULL;
		activation_step = NULL;
		reachable = NULL;
    }

    m = 0;
    n = 0;
    match_found = false;

}

// ==========================================================================================
template<class transporter_edge>
void partial_match_graph<transporter_edge>::
	extract_borders(transporter_edge& E,
    const index_t& i, const index_t& j, graph_direction dir)
// ==========================================================================================
{

    edge_border_iterator r_it, r_end, l_it, l_end;

    r_it = E.right_borders_begin();
    r_end = E.right_borders_end();
    l_it = E.left_borders_begin();
    l_end = E.left_borders_end();

    for (; l_it!=l_end; l_it++)
    {
        interval_borders_list.push_back(interval_border(*l_it,LEFT_BORDER,i,j,dir));
    }

    for (; r_it!=r_end; r_it++)
    {
        interval_borders_list.push_back(interval_border(*r_it,RIGHT_BORDER,i,j,dir));
    }

}

// ==========================================================================================
template<class transporter_edge>
void partial_match_graph<transporter_edge>::
	sort_borders()
// ==========================================================================================
{
    
    // assign sufficient space to border_intervals
    index_t ii = interval_borders_list.size();
    interval_borders = border_array(interval_borders_list.size());

    // copy from border_interval_list to border_intervals
    //std::copy(
    //    interval_borders_list.begin(),interval_borders_list.end(),
    //    interval_borders.begin());

    border_list::iterator b_it = interval_borders_list.begin();
    border_list::iterator b_end = interval_borders_list.end();

    for (; b_it!=b_end; b_it++)
    {
        interval_border border = *b_it;
        if (!(border.m_value.s<-1. || border.m_value.s>1.))
            interval_borders.push_back(border);
    }

    // sort border_intervals
    std::sort(interval_borders.begin(),interval_borders.end());

    // the unsorted list is needed no mo'.
    interval_borders_list.clear();

}

// ==========================================================================================
template<class transporter_edge>
bool partial_match_graph<transporter_edge>::
    update(const interval_border& ib)
// ==========================================================================================
{

	index_t i = ib.col();
    index_t j = ib.row();
    match_pos = ib.val();

    // update suitability-data of currently active intervals
    if (ib.type()==LEFT_BORDER)
    {
        columns[ib.col()]++;
        rows[ib.row()]++;
        if (rows[ib.row()]==1)
            active_rows++;
        if (columns[ib.col()]==1)
            active_columns++;
	    // toggle edge activation
	    edge_activated[i][j][ib.dir()] = true;
    } 

    // determine whether the active configuration of active edges is suitable
    // using the suitability information that has just been updated
    bool suitable;
    if (active_rows>=n && 
        ib.type()==RIGHT_BORDER && prev_border_type==LEFT_BORDER)
        suitable = true;
    else
        suitable = false;

    suitable = true;

    prev_border_type = ib.type();

    if (suitable)
    {            
		if (find_path())
        {
			return true;
        }
    }

    // The suitability-information for right interval broders
    // is updated after the path-update, because the border itself is still contained
    // in the intersection.
    if (ib.type()==RIGHT_BORDER)
    {
        columns[ib.col()]--;
        rows[ib.row()]--;
        if (rows[ib.row()]==0)
            active_rows--;
        if (columns[ib.col()]==0)
            active_columns--;
    	// toggle edge activation
		edge_activated[i][j][ib.dir()] = false;
    }
	
	return false;

}


// ==========================================================================================
template<class transporter_edge>
void partial_match_graph<transporter_edge>::
	build(point* P,point* Q,index_t m_P,index_t n_Q, index_t k_P,
	const arithmetic& eps, const arithmetic& delta)
// ==========================================================================================
{

	clear();

	m = m_P;
	n = n_Q;
    k = k_P;

    update_step = 0;

    // initialize the counters that help to determine the suitability
    // of the currently active intervals.
    prev_border_type = LEFT_BORDER;
    index_t i;
    columns = new index_t[m+1];
    rows = new index_t[n+1];
    for (i=0; i<=m; i++)
        columns[i] = 0;
    for (i=0; i<=n; i++)
        rows[i] = 0;
    active_columns = 0;
    active_rows = 0;

	edges = new transporter_edge**[m+1];
	edge_activated = new bool**[m+1];
	activation_step = new index_t**[m+1];
    reachable = new bool*[m+1];

	reference_points = point_pair(P[k],Q[0]);
	
	for (i=0; i<=m; i++)
	{
		edges[i] = new transporter_edge*[n+1];
		edge_activated[i] = new bool*[n+1];
		activation_step[i] = new index_t*[n+1];
        reachable[i] = new bool[n+1];

        index_t i_1 = (i+k)%(m+1);
        index_t i_2 = (i+k+1)%(m+1);

        for (index_t j=0; j<n; j++)
		{
		
            edges[i][j] = new transporter_edge[3];
			edge_activated[i][j] = new bool[3];
			activation_step[i][j] = new index_t[3];

			edges[i][j][horizontal].assign_transporter
				(point_pair(P[i_1],P[i_2]),point_pair(Q[j],Q[j]),reference_points,eps,delta);
			edges[i][j][diagonal].assign_transporter
				(point_pair(P[i_1],P[i_2]),point_pair(Q[j],Q[j+1]),reference_points,eps,delta);
			edges[i][j][vertical].assign_transporter
				(point_pair(P[i_1],P[i_1]),point_pair(Q[j],Q[j+1]),reference_points,eps,delta);
            
            // extract interval borders from the three edges just created            
            extract_borders(edges[i][j][horizontal],i,j,horizontal);
            extract_borders(edges[i][j][diagonal],i,j,diagonal);
            extract_borders(edges[i][j][vertical],i,j,vertical);

            // initially, each edge is inactive (until it is activated in an update()-step)
            edge_activated[i][j][horizontal] = false;
            edge_activated[i][j][diagonal] = false;
            edge_activated[i][j][vertical] = false;

            // activation_step 0 means that the edge has never been activated.
            activation_step[i][j][horizontal] = 0;
            activation_step[i][j][diagonal] = 0;
            activation_step[i][j][vertical] = 0;

            // since all edges are inactive, no vertex is reachable
            // (except for the source (0,0), but we'll take care of that later...).
            reachable[i][j] = false;

		}

        reachable[i][n] = false;
		edges[i][n] = new transporter_edge[3];
		edge_activated[i][n] = new bool[3];
		activation_step[i][n] = new index_t[3];

		edges[i][n][horizontal].assign_transporter
				(point_pair(P[i_1],P[i_2]),point_pair(Q[n],Q[n]),reference_points,eps,delta);
        extract_borders(edges[i][n][horizontal],i,n,horizontal);
        edge_activated[i][n][horizontal] = false;
        activation_step[i][n][horizontal] = 0;

	}

    // Before we forget: the graph's source (0,0) is always reachable!
    reachable[0][0] = true;

    // sort the extracted interval borders
    sort_borders();
		
}

// ==========================================================================================
template<class transporter_edge>
bool partial_match_graph<transporter_edge>::
	find_path()
// ==========================================================================================
{
	bool* last_column;
	bool* curr_column;

    // use dynamic programming for updating the graph.	
	bool path_found;
    match_found = false;

    bool b_v, b_h, b_d;

	// initialize last_row
	last_column = reachable[0];

	last_column[0] = true;
	for (index_t j=1; j<=n; j++)
	{
		b_v = edge_activated[0][j-1][vertical];
		last_column[j] = last_column[j-1] && b_v;
	}

	// in case m==0, curr_column[m][n] needs to specified...
	curr_column = last_column;
	index_t i=1;

	// dynamic programming
	path_found = true;
	for (; i<=m && path_found; i++)
	{
		index_t prev_col = i-1;//(i+k+m)%(m+1);
		index_t col = i;//(i+k)%(m+1);
		last_column = reachable[prev_col];
		curr_column = reachable[col];

		path_found = false;
		curr_column[0] = false;
		b_h = edge_activated[prev_col][0][horizontal];
		if (last_column[0] && b_h)
		{
			curr_column[0] = true;
			path_found = true;
		}
		for (index_t j=1; j<=n; j++)
		{
			curr_column[j] = false;

			// if-clause slightly optimized "by hand"
			b_d = edge_activated[prev_col][j-1][diagonal];                    
			if (last_column[j-1] && b_d)
			{
				curr_column[j] = true;
				path_found = true;
			} else {
				b_v = edge_activated[col][j-1][vertical];
				if (curr_column[j-1] && b_v)
				{
					curr_column[j] = true;
					path_found = true;
				} else {
					b_h = edge_activated[prev_col][j][horizontal];
    				if (last_column[j] && b_h)
					{
						curr_column[j] = true;
						path_found = true;
					}
				}
			}
		}
		if (curr_column[n])
		{
			match_found = true;
            match_column = i;
			return true;
		}
	}

	return false;

}


// ==========================================================================================
template<class transporter_edge>
partial_match_graph<transporter_edge>::motion 
    partial_match_graph<transporter_edge>::get_match()
// ==========================================================================================
{

    if (m==0 && n==0)
    {
        transporter_edge te;
        te.assign_transporter
            (point_pair(reference_points.first,reference_points.first),
             point_pair(reference_points.second,reference_points.second),
             reference_points,1.0,1.0);
        return transporter_edge::compute_match(*(te.left_borders_begin()),reference_points);

    }

    if (match_found)
    {
        return transporter_edge::compute_match(match_pos,reference_points);
    }
    motion match_motion;
    return match_motion;

}

// ==========================================================================================
template<class transporter_edge>
bool partial_match_graph<transporter_edge>::get_reparametrizations
    (partial_match_graph<transporter_edge>::discrete_reparametrization_pair& kappa_lambda)
// ==========================================================================================
{

    kappa_lambda.clear();
    index_t i = match_column;
    index_t j = n;
    index_t loop_count = 0;
    kappa_lambda.push_front(discrete_free_space_point((i+k+m)%(m+1),j));

    while (i>0 && j>0 && loop_count<=m)
    {
        index_t i_1 = i-1;//(i+m)%(m+1);
        index_t i_0 = i;//(i+m+1)%(m+1);
        if (j>0 && reachable[i_1][j-1] && edge_activated[i_1][j-1][diagonal]) {
            // activating the following lines inserts a loop for the diagonal edge...
			// discrete_free_space_point x(i,j);
            //kappa_lambda.push_front(x);
            i--;
            j--;
			discrete_free_space_point y((i+k+m)%(m+1),j);
            kappa_lambda.push_front(y);
        } else if (i>0 && reachable[i_1][j] && edge_activated[i_1][j][horizontal]) {
            i--;
			discrete_free_space_point x((i+k+m)%(m+1),j);
            kappa_lambda.push_front(x);
        } else if (j>0 && reachable[i_0][j-1] && edge_activated[i_0][j-1][vertical]) {
            j--;
			discrete_free_space_point x((i+k+m)%(m+1),j);
            kappa_lambda.push_front(x);
        } else {
			return false;
		}
        loop_count++;
    }
    
    return true;

}


// end of namespace rescue
}


#endif
