#ifndef RESCUE__PARTIAL_SAMPLING_MATCHER_H
#define RESCUE__PARTIAL_SAMPLING_MATCHER_H

#include "polygonal_curve.h"
#include "sampling_matcher.h"
#include "rescutil.h"

/** All classes, methods and variables supplied by ReSCue are implemented
within the namespace \c rescue.

\mainpage
\c ReSCue is a short form for \em Retrieval \em System \em for \em Curves. It supplies
classes and methods for handling polygonal curves, in particular matching polygonal
curves with respect to the Frechet distance.

*/
namespace rescue
{

/** A \c partial_sampling_matcher allows to approximately match two polygonal curves with respect
to the Frechet distance by approximately solving the emptyness problem for the set of all 
Frechet matches.

The algorithm used is based on \c delta - samplings of polygonal curves.

The template parameters have the following meaning:

\param arithmetic determines the data type that is used for all arithmetic.
All computations only use multiplication, division, addition, subtraction and computation
of roots. Thus, the Leda-type \c real can be used as well as \c float and \c double.

\param transporter_edge determines the objects that are assigned to edges of the
free space graph, and thus the group under which the matching problem is solved.
Each edge is assigned a (projected) transporter subset of the transformation group,
and the class \c transporter_edge needs to supply the neccessary operations for these
transporter subsets. Currently, \c circular_arc_edge can be used for matching under
rigid motions. The implementation of \c interval_edge for matching under homthetic motions
is planned.
*/

// ==========================================================================================
// ==========================================================================================
template<class fs_graph>
class partial_sampling_matcher : public sampling_matcher<fs_graph>
{

public:

    partial_sampling_matcher() 
        : sampling_matcher() {};

    partial_sampling_matcher(curve& P, curve& Q, const arithmetic& ee, const arithmetic& dd)
        : sampling_matcher<fs_graph>(P,Q,ee,dd) {};

    partial_sampling_matcher(curve& P, curve& Q, const arithmetic& ee)
        : sampling_matcher<fs_graph>(P,Q,ee) {};

    partial_sampling_matcher(curve& P, curve& Q, const int max_m, const int max_n, const arithmetic& ee, const arithmetic& dd)
        : sampling_matcher<fs_graph>(P,Q,max_m,max_n,ee,dd) {};

    partial_sampling_matcher(curve& P, curve& Q, const int max_m, const int max_n, const arithmetic& ee)
        : sampling_matcher<fs_graph>(P,Q,max_m,max_n,ee) {};


    ~partial_sampling_matcher();

    /** Runs the matching algorithm, taking O(m^3n^2) time if both \c P and \c Q are
    \c delta - sampled.
    */
    bool match();

    /** Deallocates all memory used by the \c partial_sampling_matcher object.
    */
    void clear();

};
// ==========================================================================================
// ==========================================================================================

// ==========================================================================================
template<class fs_graph>
bool partial_sampling_matcher<fs_graph>::match()
// ==========================================================================================
{

    if (m<0 || n<0)
        return false;

    Gamma.clear();

    for (index_t k=0; k<=m; k++)
    {
        
	    Gamma.build(P,Q,m,n,k,eps,delta);

        // We treat the case that both P and Q are points as an exception
        if (m==0 && n==0)
        {
            match_motion = Gamma.get_match();
            return true;
        }

	    fs_graph::border_iterator b_it = Gamma.border_begin();
	    fs_graph::border_iterator b_end = Gamma.border_end();

	    for (; b_it!=b_end; b_it++)
	    {

		    if (Gamma.update(*b_it))
            {
                match_motion = Gamma.get_match();
                parametrizations_available = Gamma.get_reparametrizations(match_kappa_lambda);
			    return true;
            }

	    }
    }

    Gamma.clear();

    return false;

}


// ==========================================================================================
template<class fs_graph>
void partial_sampling_matcher<fs_graph>::clear()
// ==========================================================================================
{
    sampling_matcher<fs_graph>::clear();
}
    
// ==========================================================================================
template<class fs_graph>
partial_sampling_matcher<fs_graph>::~partial_sampling_matcher()
// ==========================================================================================
{
    clear();
}
    



// ==========================================================================================
// ==========================================================================================

// end of namespace rescue
}

#endif
