#ifndef RIGIDMOTION_H
#define RIGIDMOTION_H

#include <iostream>
#include "angle.h"

namespace rescue
{

// ==========================================================================================
// ==========================================================================================

template <class arithmetic>
class rigid_motion
{

public:    

    typedef unit_circle_angle<arithmetic> angle_t;

    arithmetic t_x;
    arithmetic t_y;

    arithmetic sin_phi; // sin und cos des Drehwinkels
    arithmetic cos_phi;

    inline rigid_motion<arithmetic>& operator*=(const rigid_motion<arithmetic>& M);
    inline rigid_motion<arithmetic> operator*(const rigid_motion<arithmetic>& M) const;
    inline rigid_motion<arithmetic> operator!() const;

    rigid_motion<arithmetic>();
    rigid_motion<arithmetic>(const arithmetic&, const arithmetic& , const arithmetic&, const arithmetic&);
    rigid_motion<arithmetic>(const angle_t&, const arithmetic&, const arithmetic&);
    rigid_motion<arithmetic>(const arithmetic&, const arithmetic&);
    rigid_motion<arithmetic>(const angle_t&);

    static rigid_motion<arithmetic> one();

    angle_t angle() const;

    arithmetic operator[](int) const;
};

// ==========================================================================================
// ==========================================================================================


// ==========================================================================================
template <class arithmetic>
inline rigid_motion<arithmetic>& rigid_motion<arithmetic>::operator*=(const rigid_motion<arithmetic>& M)
// ==========================================================================================
{
    
    arithmetic c = cos_phi * M.cos_phi - sin_phi * M.sin_phi;
    arithmetic s = sin_phi * M.cos_phi + cos_phi * M.sin_phi;

    t_x += cos_phi*M.t_x - sin_phi*M.t_y;
    t_y += sin_phi*M.t_x + cos_phi*M.t_y;

    sin_phi = s;
    cos_phi = c;

    return *this;
}

// ==========================================================================================
template <class arithmetic>
inline rigid_motion<arithmetic> rigid_motion<arithmetic>::operator*(const rigid_motion<arithmetic>& M) const
// ==========================================================================================
{
    rigid_motion<arithmetic> r = *this;
    r *= M;
    return r;
}

// ==========================================================================================
template <class arithmetic>
inline rigid_motion<arithmetic> rigid_motion<arithmetic>::operator!() const
// ==========================================================================================
{
    rigid_motion<arithmetic> r;

    r.cos_phi = cos_phi;
    r.sin_phi = -sin_phi;

    r.t_x = -t_x*cos_phi - t_y*sin_phi;
    r.t_y = t_x*sin_phi - t_y*cos_phi;

    return r;
}

// ==========================================================================================
template <class arithmetic>
rigid_motion<arithmetic>::rigid_motion()
// ==========================================================================================
{
    sin_phi = 0.0;
    cos_phi = 1.0;
    t_x = 0.0;    
    t_y = 0.0;    
}

// ==========================================================================================
template <class arithmetic>
rigid_motion<arithmetic>::rigid_motion(const arithmetic& s_phi, const arithmetic& c_phi, const arithmetic& x, const arithmetic& y)
// ==========================================================================================
{
    sin_phi = s_phi;
    cos_phi = c_phi;
    t_x = x;    
    t_y = y;    
}

// ==========================================================================================
template <class arithmetic>
rigid_motion<arithmetic>::rigid_motion(const rigid_motion<arithmetic>::angle_t& a, const arithmetic& x, const arithmetic& y)
// ==========================================================================================
{
    sin_phi = a.sin();
    cos_phi = a.vos();
    t_x = x;    
    t_y = y;    
}

// ==========================================================================================
template <class arithmetic>
rigid_motion<arithmetic>::rigid_motion(const arithmetic& x, const arithmetic& y)
// ==========================================================================================
{
    sin_phi = 0.0;
    cos_phi = 1.0;
    t_x = x;    
    t_y = y;    
}

// ==========================================================================================
template <class arithmetic>
rigid_motion<arithmetic>::rigid_motion(const rigid_motion<arithmetic>::angle_t& a)
// ==========================================================================================
{
    sin_phi = a.sin();
    cos_phi = a.cos();
    t_x = 0.0;    
    t_y = 0.0;    
}

// ==========================================================================================
template <class arithmetic>
rigid_motion<arithmetic>::angle_t rigid_motion<arithmetic>::angle() const
// ==========================================================================================
{
    return angle_t(sin_phi,cos_phi);
}

// ==========================================================================================
template <class arithmetic>
rigid_motion<arithmetic> rigid_motion<arithmetic>::one()
// ==========================================================================================
{
    rigid_motion<arithmetic> o(0.0,1.0,0.0,0.0);
    return o;
}

// ==========================================================================================
template <class arithmetic>
arithmetic rigid_motion<arithmetic>::operator[](int i) const
// ==========================================================================================
{
    
    switch(i)
    {
    case 0:
        return t_x;
        break;
    case 1:
        return t_y;
        break;
    case 2:
        return angle().rad();
        break;
    }
    return 0.0;

}

// ==========================================================================================
template <class arithmetic>
std::ostream& operator<<(std::ostream& o, const rigid_motion<arithmetic>& rm)
// ==========================================================================================
{
    double d_phi = (atan(rm.sin_phi/rm.cos_phi));
    if (rm.sin_phi<0.0 && rm.cos_phi>0.0)
        d_phi += 2.0*PI;
    else if (rm.sin_phi>0.0 && rm.cos_phi<0.0)
        d_phi += PI;
    else if (rm.sin_phi<0.0 && rm.cos_phi<0.0)
        d_phi += PI;
    double d_x = rm.t_x;
    double d_y = rm.t_y;
    o<<"("<<d_phi<<","<<d_x<<","<<d_y<<")";

    return o;
}


}

#endif


