#ifndef SAMPLING_MATCHER_H
#define SAMPLING_MATCHER_H

#include "polygonal_curve.h"
#include "rescutil.h"

/** All classes, methods and variables supplied by ReSCue are implemented
within the namespace \c rescue.

\mainpage
\c ReSCue is a short form for \em Retrieval \em System \em for \em Curves. It supplies
classes and methods for handling polygonal curves, in particular matching polygonal
curves with respect to the Frechet distance.

*/
namespace rescue
{

/** A \c sampling_matcher allows to approximately match two polygonal curves with respect
to the Frechet distance by approximately solving the emptyness problem for the set of all 
Frechet matches.

The algorithm used is based on \c delta - samplings of polygonal curves.

The template parameters have the following meaning:

\param arithmetic determines the data type that is used for all arithmetic.
All computations only use multiplication, division, addition, subtraction and computation
of roots. Thus, the Leda-type \c real can be used as well as \c float and \c double.

\param transporter_edge determines the objects that are assigned to edges of the
free space graph, and thus the group under which the matching problem is solved.
Each edge is assigned a (projected) transporter subset of the transformation group,
and the class \c transporter_edge needs to supply the neccessary operations for these
transporter subsets. Currently, \c circular_arc_edge can be used for matching under
rigid motions. The implementation of \c interval_edge for matching under homthetic motions
is planned.
*/

// ==========================================================================================
// ==========================================================================================
template<class fs_graph>
class sampling_matcher
{

public:

    typedef fs_graph::transporter_edge_class transporter_edge;
    typedef fs_graph::motion motion;
    typedef fs_graph::arithmetic_t arithmetic;
    typedef arithmetic arithmetic_t;
    typedef fs_graph::index_t index_t;
    typedef fs_graph::point point;
	typedef fs_graph::discrete_reparametrization_pair discrete_reparametrization_pair;

    typedef polygonal_curve<point> curve;
    typedef curve::vertex vertex;
    typedef curve::edge edge;

protected:
    
    /** Length of the \c delta - sampled polygonal curve \c P.
    */
    index_t m;

    /** Length of \c delta - sampled the polygonal curve \c Q.
    */
    index_t n;

    /** Reparemtrization of \c P indicating a match. Currently unused.
    */
    index_t* kappa;

    /** Reparemtrization of \c Q indicating a match. Currently unused.
    */
    index_t* lambda;
    
    /** \c delta - sampled Polygonal curve of length \c m. \c P[0], \c P[1], \c ..., \c P[m] are
    the vertices of \c P.
    */
    point* P;

    /** \c delta - sampled Polygonal curve of length \c n. \c Q[0], \c Q[1], \c ..., \c Q[n] are
    the vertices of \c Q.
    */
    point* Q;

    /** Fault tolerance for matching.
    */
    arithmetic eps; 

    /** Sampling density for matching; controls the aulity of approximation.
    */
    arithmetic delta;

    /** Determines whether \c match() has been called since initialization or since
    the last call of \c clear().
    */
    bool matched;

    /** Indicates whether the last call of \c match() yielded a match for \c P and \c Q.
    Only specified if \c matched is \c true.
    */
    bool match_found;

	/** Indicates whether parametrizations computed during the last call of \c match()
	can be specified.
	*/
	bool parametrizations_available;
    
	/** The free space graph used for matching.
    */
	fs_graph Gamma;

    /** A match found by \c match(). Only specified if both \c matched and \c match_found
    are \c true.
    */
    motion match_motion;

	
    /** A reparametrization found by \c match(). Only specified if both \c matched and 
	\c match_found are \c true.
    */
	discrete_reparametrization_pair match_kappa_lambda;

    /** Computes the Euclidean distance between \c x and \c y.
    */
    static inline arithmetic dist(const point& x, const point& y);

public:

    /** Returns a pair (\c kappa, \c lambda) containing the curves \c P and \c gQ
    reparemtrized by \c kappa and \c lambda, where \c g is a group element found
    by \c find_path() (if such \c g is available).
    */
	bool get_reparametrizations(discrete_reparametrization_pair&);


    /** Initializes an object that computes a solution to the emptyness-problem
    of discrete Frechet-Matches. \c ee specifes the fault tolerance used for the 
    matching task.
    */
    sampling_matcher(curve& P, curve& Q, const arithmetic& ee);
    
    /** Initializes an object that computes a solution to the emptyness-problem
    of Frechet-Matches. \c ee specifes the fault tolerance and \c dd the sampling density
    (and thus the quality of approximation) used for the matching task.
    */
    sampling_matcher(curve& cP, curve& cQ, const arithmetic& ee, const arithmetic& dd);

    /** Default constructor.
    */
    sampling_matcher();

    /** Default destructor.
    */
    ~sampling_matcher();

    /** Runs the matching algorithm, taking O(m^2n^2) time if both \c P and \c Q are
    \c delta - sampled.
    */
    bool match();

    /** Returns a transformation computed by \c match(). Only specified if \c match() has 
    been called before and returned \c true.
    */
    motion get_match();

    /** Returns a pair (\c P_kappa, \c P_lambda) containing the curves \c P and \c gQ
    reparemtrized by \c kappa and \c lambda, where \c g is a group element found
    by \c find_path() (if such \c g is available).
    */
    bool get_reparametrized_curves(curve& P_kappa, curve& Q_lambda);

    /** Deallocates all memory used by the \c sampling_matcher object.
    */
    void clear();

};
// ==========================================================================================
// ==========================================================================================



// ==========================================================================================
template<class fs_graph>
sampling_matcher<fs_graph>::sampling_matcher()
// ==========================================================================================
{

    m = 0;
    n = 0;

    kappa = NULL;
    lambda = NULL;
    
    P = NULL;
    Q = NULL;

    eps = 0.0;
    delta = 0.0;

	parametrizations_available = false;
	matched = false;
	match_found = false;
}


// ==========================================================================================
template<class fs_graph>
sampling_matcher<fs_graph>::sampling_matcher(curve& cP, curve& cQ, const arithmetic& ee, const arithmetic& dd)
// ==========================================================================================
{

	parametrizations_available = false;
	matched = false;
	match_found = false;

    if (cP.length()<0 || cQ.length()<0)
    {
        // matching curves without vertices does not make sense...
        m = -1;
        n = -1;

        kappa = NULL;
        lambda = NULL;
    
        P = NULL;
        Q = NULL;

        eps = 0.0;
        delta = 0.0;

        return;
    }

    eps = ee;
    delta = dd;
    
    kappa = new index_t[m+n+2];
    lambda = new index_t[m+n+2];

    // Construct delta-Samplings

	curve::vertex_iterator v_it, v_end;
    v_it = cP.vertex_begin();
    v_end = cP.vertex_end();
	
	std::list<point> P_cp;
	point P0,P1;
	P1 = (*v_it).pos;
    P_cp.push_back(P1);
	v_it++;

	for (; v_it!=v_end; v_it++)
    {

        // sample the line segment of <p_i,p_{i+1}>
        P0 = P1;
        P1 = (*v_it).pos;

        arithmetic length = sqrt((P0.x-P1.x)*(P0.x-P1.x)+(P0.y-P1.y)*(P0.y-P1.y));
        arithmetic Dx = 2.0*delta*(P1.x-P0.x)/length;
        arithmetic Dy = 2.0*delta*(P1.y-P0.y)/length;

        arithmetic covered = 2.0*delta;

        while (covered<=length)
        {
            P0.x += Dx;
            P0.y += Dy;
            P_cp.push_back(P0);
            covered += 2.0*delta;
        }

        P_cp.push_back(P1);        

    }

    m = P_cp.size()-1;
    P = new point[P_cp.size()];
    std::list<point>::iterator p_it = P_cp.begin();
    index_t i=0;

    for (; p_it!=P_cp.end(); p_it++)
    {
        P[i]=*p_it;
        i++;
    }

    std::list<point> Q_cp;
    v_it = cQ.vertex_begin();
    v_end = cQ.vertex_end();
	point Q0,Q1;
	Q1 = (*v_it).pos;
    Q_cp.push_back(Q1);
	v_it++;

	for (; v_it!=v_end; v_it++)
    {

        // sample the line segment of <p_i,p_{i+1}>
        Q0 = Q1;
        Q1 = (*v_it).pos;

        arithmetic length = sqrt((Q0.x-Q1.x)*(Q0.x-Q1.x)+(Q0.y-Q1.y)*(Q0.y-Q1.y));
        arithmetic Dx = 2.0*delta*(Q1.x-Q0.x)/length;
        arithmetic Dy = 2.0*delta*(Q1.y-Q0.y)/length;

        arithmetic covered = 2.0*delta;

        while (covered<=length)
        {
            Q0.x += Dx;
            Q0.y += Dy;
            Q_cp.push_back(Q0);
            covered += 2.0*delta;
        }

        Q_cp.push_back(Q1);        

    }

    n = Q_cp.size()-1;
    Q = new point[Q_cp.size()];
    std::list<point>::iterator q_it = Q_cp.begin();
    index_t j=0;

    for (; q_it!=Q_cp.end(); q_it++)
    {
        Q[j]=*q_it;
        j++;
    }

}


// ==========================================================================================
template<class fs_graph>
sampling_matcher<fs_graph>::sampling_matcher(curve& cP, curve& cQ, const arithmetic& ee)
// ==========================================================================================
{

	parametrizations_available = false;
	matched = false;
	match_found = false;

    if (cP.length()<0 || cQ.length()<0)
    {
        // matching curves without vertices does not make sense...
        m = -1;
        n = -1;

        kappa = NULL;
        lambda = NULL;
    
        P = NULL;
        Q = NULL;

        eps = 0.0;
        delta = 0.0;

        return;
    }

    eps = ee;
    delta = 0.0;
    
    kappa = new index_t[m+n+2];
    lambda = new index_t[m+n+2];

    m = cP.length();
    P = new point[cP.length()+1];
    Curve::vertex_iterator p_it = cP.vertex_begin();
    index_t i=0;

    for (; p_it!=cP.vertex_end(); p_it++)
    {
        P[i]=(*p_it).pos;
        i++;
    }

    n = cQ.length();
    Q = new point[cQ.length()+1];
    Curve::vertex_iterator q_it = cQ.vertex_begin();
    i=0;

    for (; q_it!=cQ.vertex_end(); q_it++)
    {
        Q[i]=(*q_it).pos;
        i++;
    }

}

// ==========================================================================================
template<class fs_graph>
bool sampling_matcher<fs_graph>::match()
// ==========================================================================================
{

    if (m<0 || n<0)
        return false;

	Gamma.build(P,Q,m,n,eps,delta);

    // We treat the case that both P and Q are points as an exception
    if (m==0 && n==0)
    {
        match_motion = Gamma.get_match();
        return true;
    }

	fs_graph::border_iterator b_it = Gamma.border_begin();
	fs_graph::border_iterator b_end = Gamma.border_end();

    int upd_cnt = 0;
	for (; b_it!=b_end; b_it++)
	{
        upd_cnt++;

		if (Gamma.update(*b_it))
        {
            match_motion = Gamma.get_match();
            parametrizations_available = Gamma.get_reparametrizations(match_kappa_lambda);
			return true;
        }

	}

    Gamma.clear();

    return false;

}


// ==========================================================================================
template<class fs_graph>
sampling_matcher<fs_graph>::motion sampling_matcher<fs_graph>::get_match()
// ==========================================================================================
{
    return match_motion;
}

// ==========================================================================================
template<class fs_graph>
void sampling_matcher<fs_graph>::clear()
// ==========================================================================================
{
    if (P!=NULL)
        delete[] P;
    if (Q!=NULL)
        delete[] Q;
    if (kappa!=NULL)
        delete[] kappa;
    if (lambda!=NULL)
        delete[] lambda;

    P = NULL;
    Q = NULL;
    kappa = NULL;
    lambda = NULL;

    m = 0;
    n = 0;

    eps = 0.0;
    delta = 0.0;

    Gamma.clear();

}
    
// ==========================================================================================
template<class fs_graph>
sampling_matcher<fs_graph>::~sampling_matcher()
// ==========================================================================================
{
    clear();
}
    

// ==========================================================================================
template<class fs_graph>
inline sampling_matcher<fs_graph>::arithmetic sampling_matcher<fs_graph>::dist(const point& P, const point& Q)
// ==========================================================================================
{
    return sqrt(sqr<arithmetic>(P.x-Q.x)+sqr<arithmetic>(P.y-Q.y));
}

// ==========================================================================================
template<class fs_graph>
bool sampling_matcher<fs_graph>::get_reparametrizations(sampling_matcher<fs_graph>::discrete_reparametrization_pair& d)
// ==========================================================================================
{
	if (!reparametrizations_available)
		return false;
    kappa_lambda = match_kappa_lambda;
	return true;
}


// ==========================================================================================
template<class fs_graph>
bool sampling_matcher<fs_graph>::
    get_reparametrized_curves(curve& P_kappa, curve& Q_lambda)
// ==========================================================================================
{
    P_kappa.clear();
    Q_lambda.clear();

    if (parametrizations_available==false)
        return false;

    typename fs_graph::discrete_reparametrization_pair::iterator kl_pos = match_kappa_lambda.begin();
    typename fs_graph::discrete_reparametrization_pair::iterator kl_end = match_kappa_lambda.end();

    for (; kl_pos!=kl_end; kl_pos++)
    {
        P_kappa.push_back(new polygon_vertex<point>(P[kl_pos->first]));
        point gQ = Q[kl_pos->second];
        gQ *= match_motion;
        Q_lambda.push_back(new polygon_vertex<point>(gQ));
    }

    return true;

}



// ==========================================================================================
// ==========================================================================================

// end of namespace rescue
}

#endif
